package de.mdelab.mlsdm.interpreter.searchModel;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.MLSDMInterpreter;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMReferenceIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMCallActionsExpressionAnalyzer;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMICLExpressionAnalyzer;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMOCLExpressionAnalyzer;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMStrategyFactory;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.StoryPatternMatcher;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * This is a parametrizable implementation of the MLSDM interpreter which
 * uses a search model based pattern matcher as its default pattern matcher.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMSearchModelBasedInterpreter extends MLSDMInterpreter {

	/*
	 * Constant for enabling and disabling the creation and usage of a
	 * reference adapter for optimized edge checks, reverse edge traversal,
	 * ...
	 */
	public static String ENABLE_REFERENCE_ADAPTER = "ENABLE_REFERENCE_ADAPTER";
	
	/*
	 * Constant for enabling and disabling domain matching. Note that this
	 * requires a reference adapter.
	 */
	public static String ENABLE_DOMAIN_MATCHING = "ENABLE_DOMAIN_MATCHING";
	
	/*
	 * Constant for enabling and disabling staged matching when using staged
	 * indices.
	 */
	public static String ENABLE_STAGED_INDEX_MATCHING = "ENABLE_STAGED_INDEX_MATCHING";
	
	/*
	 * Constant for providing a specific reference adapter to be used by the
	 * interpreter.
	 */
	public static String REFERENCE_ADAPTER = "REFERENCE_ADAPTER";
	
	/*
	 * Constant for assigning specific matching strategies to certain story nodes.
	 */
	public static String STRATEGY_FACTORY = "STRATEGY_FACTORY";
	
	/*
	 * Map of interpreter parameters to steer the interpreter's behavior.
	 */
	protected Map<String, Object> interpreterParameters;
	
	/*
	 * Reference adapter to be used by all pattern matchers of this interpreter.
	 */
	protected MLSDMReferenceIndex referenceAdapter;
	
	/*
	 * Expression analyzer to be used by all pattern matchers of this
	 * interpreter.
	 */
	protected ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager;
	
	public MLSDMSearchModelBasedInterpreter(
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager) {
		this(expressionInterpreterManager, Collections.<String, Object>emptyMap());
	}

	public MLSDMSearchModelBasedInterpreter(
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			Map<String, Object> interpreterParameters) {
		this(expressionInterpreterManager, MLSDMMetamodelFacadeFactory.INSTANCE, interpreterParameters);
	}
	
	public MLSDMSearchModelBasedInterpreter(
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			MLSDMMetamodelFacadeFactory p_facadeFactory, Map<String, Object> interpreterParameters) {
		super( expressionInterpreterManager, p_facadeFactory);

		this.interpreterParameters = interpreterParameters;
		
		expressionAnalyzerManager = new ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression>(MLSDMMetamodelFacadeFactory.INSTANCE.getExpressionFacade());
		expressionAnalyzerManager.registerExpressionAnalyzer("OCL", new MLSDMOCLExpressionAnalyzer(expressionAnalyzerManager));
		expressionAnalyzerManager.registerExpressionAnalyzer("ICL", new MLSDMICLExpressionAnalyzer(expressionAnalyzerManager));
		expressionAnalyzerManager.registerExpressionAnalyzer("CallActions", new MLSDMCallActionsExpressionAnalyzer(expressionAnalyzerManager));
		
		if(interpreterParameters.containsKey(REFERENCE_ADAPTER)) {
			referenceAdapter = (MLSDMReferenceIndex) interpreterParameters.get(REFERENCE_ADAPTER);
			getNotificationEmitter().addNotificationReceiver(referenceAdapter);
		}
		else if(!interpreterParameters.containsKey(ENABLE_REFERENCE_ADAPTER) || (Boolean) interpreterParameters.get(ENABLE_REFERENCE_ADAPTER)) {
			referenceAdapter = new MLSDMReferenceIndex();
			getNotificationEmitter().addNotificationReceiver(referenceAdapter);
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.SDMInterpreter#executeActivity(java.lang.Object, java.util.Collection)
	 */
	@Override
	public Map<String, Variable<EClassifier>> executeActivity(final Activity activity, final Collection<Variable<EClassifier>> parameters) throws SDMException{
		if(referenceAdapter != null) {
			for(Variable<EClassifier> parameter:parameters) {
				if(parameter.getValue() instanceof EObject) {
					referenceAdapter.registerEObject((EObject) parameter.getValue());
				}
			}
		}
		
		return super.executeActivity(activity, parameters);
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.mlsdm.interpreter.MLSDMInterpreter#createStoryPatternMatcher(de.mdelab.mlstorypatterns.StoryPattern, de.mdelab.sdm.interpreter.core.variables.VariablesScope)
	 */
	@Override
	protected StoryPatternMatcher<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createStoryPatternMatcher(
			StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope)
			throws SDMException {		
		if(interpreterParameters.containsKey(STRATEGY_FACTORY)) {
			MLSDMStrategyFactory strategyFactory = (MLSDMStrategyFactory) interpreterParameters.get(STRATEGY_FACTORY);
			return new MLSDMSearchModelBasedMatcher(storyPattern,
					variablesScope,
					getFacadeFactory(),
					getExpressionInterpreterManager(),
					expressionAnalyzerManager,
					strategyFactory,
					referenceAdapter,
					getNotificationEmitter(),
					interpreterParameters);
		}
		else {
			return new MLSDMSearchModelBasedMatcher(storyPattern,
					variablesScope,
					getFacadeFactory(),
					getExpressionInterpreterManager(),
					expressionAnalyzerManager,
					referenceAdapter,
					getNotificationEmitter(),
					interpreterParameters);
		}
	}
	
	/**
	 * Get the first successful matching order used to find a match of a
	 * given story node.
	 * 
	 * @param storyPattern
	 * 			The story node for which to retrieve the matching order.
	 * @return	A list containing the matching actions executed in the
	 * 			course of the first successful graph search in order of
	 * 			execution.
	 */
	public List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> getMatchingOrder(StoryPattern storyPattern) {
		if((MLSDMSearchModelBasedMatcher) getStoryPatternMatchers().get(storyPattern) != null){
			return ((MLSDMSearchModelBasedMatcher) getStoryPatternMatchers().get(storyPattern)).getCurrentMatchingOrder();
		}
		else {
			return Collections.emptyList();
		}
	}

	public MLSDMReferenceIndex getReferenceAdapter() {
		return referenceAdapter;
	}

}
