package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the container node
 * of a containment link pattern constraint. It can be executed
 * once a mapping has been created for the contained node.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMContainerMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specifically for a container matching
	 * action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMContainerMatchingActionMatchState extends MatchState{
		
		/*
		 * The last candidate mapping generated for the container
		 * node.
		 */
		private EObject previousContainer;
		
		private MLSDMContainerMatchingActionMatchState(EObject previousContainer) {
			this.setPreviousContainer(previousContainer);
		}

		public EObject getPreviousContainer() {
			return previousContainer;
		}

		public void setPreviousContainer(EObject previousContainer) {
			this.previousContainer = previousContainer;
		}
		
	}
	
	/*
	 * Containment link to match.
	 */
	private AbstractStoryPatternLink link;
	
	/*
	 * Containing node of the containment link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode;
	
	/*
	 * Container node of the containment link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode;

	public MLSDMContainerMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, AbstractStoryPatternLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode) {
		super(id, matcher);
		this.link = link;
		this.containingNode = containingNode;
		this.containedNode = containedNode;
		this.addRequiredBinding(containedNode);
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		
		if(!containedNode.isBound() || containingNode.isBound()) {
			return null;
		}
		else{
			AbstractStoryPatternObject containedSPO = containedNode.getSpo();
			EObject containedInstance = (EObject) matcher.getVariable(containedSPO.getName()).getValue();
			return new MLSDMContainerMatchingActionMatchState(containedInstance);
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!containedNode.isBound() || containingNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			//TODO estimate height of containment tree
			return 5;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		EObject containedInstance = (EObject) matcher.getVariable(containedNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), containedInstance, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		if(!(matchState instanceof MLSDMContainerMatchingActionMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMContainerMatchingActionMatchState ms = (MLSDMContainerMatchingActionMatchState) matchState;
			EObject previousContainer = ms.getPreviousContainer();
			EObject containingInstance = null;
			while(!valid && previousContainer.eContainer() != null) {
				/*
				 * Try mapping container to parent instance of previous candidate
				 * in containment tree.
				 */
				containingInstance = previousContainer.eContainer();
				valid = matcher.instanceMatchesSPO(containingInstance, containingNode.getSpo()) && matcher.checkMatchingHistory(this.getPatternConstraint(), containingInstance);
				
				previousContainer = containingInstance;
				ms.setPreviousContainer(previousContainer);
			}
			
			if(valid) {
				transaction.setTargetInstance(containingInstance);
				transaction.setSPO(containingNode.getSpo());
			}
			else {
				matcher.getNotificationEmitter().storyPatternObjectNotBound(containingNode.getSpo(), matcher.getVariablesScope(),
						matcher);
			}
			
			transaction.setValid(valid);
			transaction.addPatternNode(containingNode);

			matcher.updateMatchingHistory(this.getPatternConstraint(), containingInstance);
		}
		return transaction;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse \t (" + link.getTarget().getName() + " -*container*-> " + link.getSource().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMContainerMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containingNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containedNode));
	}
}
