package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Iterator;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching a node based
 * on a domain (type) constraint formulated over it.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMDomainMatchMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specific to the domain matching
	 * action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMDomainMatchMatchState extends MatchState {
		
		private Iterator<EObject> domainIterator;
		
		private MLSDMDomainMatchMatchState(Iterator<EObject> domainIterator) {
			this.domainIterator = domainIterator;
		}
		
	}
	
	/*
	 * Node over which the domain constraint is formulated.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode;

	public MLSDMDomainMatchMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode) {
		super(id, matcher);
		this.patternNode = patternNode;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		if(patternNode.isBound()) {
			return null;
		}
		EClass type = (EClass) patternNode.getSpo().getType();
		return new MLSDMDomainMatchMatchState(matcher.getReferenceAdapter().getDomain(type).iterator());
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(patternNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		EClass type = (EClass) patternNode.getSpo().getType();
		return matcher.getReferenceAdapter().getDomain(type).size();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		if(!(matchState instanceof MLSDMDomainMatchMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMDomainMatchMatchState ms = (MLSDMDomainMatchMatchState) matchState;
			Iterator<EObject> domainIterator = ms.domainIterator;
			Object targetInstance = null;
			while(!valid && domainIterator.hasNext()) {
				targetInstance = domainIterator.next();
				valid = matcher.instanceMatchesSPO(targetInstance, patternNode.getSpo()) && matcher.checkMatchingHistory(this.getPatternConstraint(), targetInstance);
			}
			
			if(valid) {
				transaction.setTargetInstance(targetInstance);
				transaction.setSPO(patternNode.getSpo());
			}
			else {
				matcher.getNotificationEmitter().storyPatternObjectNotBound(patternNode.getSpo(), matcher.getVariablesScope(),
						matcher);
			}
			
			transaction.setValid(valid);
			transaction.addPatternNode(patternNode);

			matcher.updateMatchingHistory(this.getPatternConstraint(), targetInstance);
		}
		return transaction;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getNode() {
		return patternNode;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "match\t\t(" + patternNode.getSpo().getName() + ": " + patternNode.getSpo().getType().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMDomainMatchMatchingAction(id, matcher, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(patternNode));
	}

}
