package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check for a constraint formulated in some expression
 * language. It can be executed once mappings have been
 * created for all nodes which the expression depends on.
 * Ideally, these are only those nodes which are referenced
 * by the expression.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMExpressionCheckMatchingAction extends MLSDMCheckMatchingAction {
	
	/*
	 * Expression representing the constraint.
	 */
	private MLExpression expression;
	
	public MLSDMExpressionCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, MLExpression expression,
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> dependencies) {
		super(id, matcher);
		this.expression = expression;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dependency:dependencies) {
			this.addRequiredBinding(dependency);
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:this.getRequiredBindings()) {
			if(!patternNode.isBound()) {
				return PatternConstraint.MATCHING_NOT_POSSIBLE;
			}
		}
		return PatternConstraint.CHECK_ONLY;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkExpression());
		}
		return transaction;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		return new MLSDMCheckOnceMatchState();
	}
	
	/**
	 * Perform the actual check of the expression using the
	 * expression interpreter manager of the pattern matcher.
	 * 
	 * @return True if the expression is evaluated to true.
	 */
	private boolean checkExpression() {
		boolean success = false;
		try {
			success = (Boolean) matcher.getExpressionInterpreterManager().evaluateExpression(expression, EcorePackage.Literals.EOBJECT, null, matcher.getVariablesScope()).getValue();
		} catch (ExpressionsInterpreterException e) {
			e.printStackTrace();
		}

		if(success) {
			matcher.getNotificationEmitter().storyPatternConstraintHolds(expression, matcher.getStoryPattern(), matcher.getVariablesScope(), matcher);
		}
		else {
			matcher.getNotificationEmitter().storyPatternConstraintViolated(expression, matcher.getStoryPattern(), matcher.getVariablesScope(), matcher);
		}
		
		return success;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkExpression();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check\t\t(" + (expression instanceof MLStringExpression ? ((MLStringExpression) expression).getExpressionString() : expression.toString()) + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requiredClones
				= new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> requirement:getRequiredBindings()) {
			requiredClones.add((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(requirement));
		}
		return new MLSDMExpressionCheckMatchingAction(id, matcher, expression, requiredClones);
	}
}
