package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the target node
 * of a link in the story pattern. It can be executed once
 * the source node of the link has been mapped.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMLinkTraverseMatchingAction extends MLSDMMatchingAction {

	/*
	 * The link.
	 */
	private StoryPatternLink link;
	
	/*
	 * The link type.
	 */
	private EStructuralFeature feature;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	
	private MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction;
	
	/**
	 * This is a match state specific to the traverse link
	 * matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private static class MLSDMTraverseLinkMatchState extends MatchState{

		/*
		 * Iterator of possible link targets in the instance model.
		 */
		private Iterator<? extends Object> featureIterator;
		
		private MLSDMTraverseLinkMatchState(Iterator<? extends Object> featureIterator) {
			this.featureIterator = featureIterator;
		}

		private Iterator<? extends Object> getFeatureIterator() {
			return featureIterator;
		}
		
	}
	
	public MLSDMLinkTraverseMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher,
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint,
			StoryPatternLink link,
			EStructuralFeature feature,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher, patternConstraint);
		this.link = link;
		this.feature = feature;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		this.addRequiredBinding(sourceNode);
		initializeTransaction();
	}

	/**
	 * Create and initialize a matching transaction for this matching action.
	 * The created transaction can be reused to avoid overhead since for each
	 * matching action, only one transaction can be on the matching stack at
	 * a time.
	 */
	private void initializeTransaction() {
		transaction = new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(patternConstraint, matcher);

		transaction.setSPO(targetNode.getSpo());
		transaction.addPatternNode(targetNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected double doEstimateCardinality()  {
		if(sourceNode.isBound() && !targetNode.isBound()) {
			if(!feature.isMany()) {
				return 1;
			}
			else {
				AbstractStoryPatternObject sourceSPO = sourceNode.getSpo();
			
				Variable<EClassifier> sourceVariable = matcher.getVariablesScope().getVariable(sourceSPO.getName());
				
				//TODO should not work if this actually is reverse traversal...
				switch(link.getLinkPositionConstraint()) {
					case BETWEEN :
						return Math.max(1, ((Collection<Object>)((EObject)sourceVariable.getValue()).eGet(feature)).size() - 2);
					case FIRST :
						return 1;
					case LAST :
						return 1;
					case UNCONSTRAINED :
						return ((Collection<Object>)((EObject)sourceVariable.getValue()).eGet(feature)).size();
					default :
						return ((Collection<Object>)((EObject)sourceVariable.getValue()).eGet(feature)).size();
				}
			}
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		AbstractStoryPatternObject sourceSPO = sourceNode.getSpo();
		
		EObject instanceObject = (EObject) matcher.getVariable(sourceSPO.getName()).getValue();
		Iterator<? extends Object> instanceLinkIterator;
		if(!feature.isMany()) {
			List<Object> singletonList = new ArrayList<Object>();
			Object target = instanceObject.eGet(feature);
			if(target != null) {
				singletonList.add(target);
			}
			instanceLinkIterator = singletonList.iterator();
		}
		else {
			List<?> instanceLinks = ((List<?>)instanceObject.eGet(feature));
			switch(link.getLinkPositionConstraint()) {
				case BETWEEN :
					instanceLinkIterator = instanceLinks.subList(1, instanceLinks.size() - 1).iterator(); 
					break;
				case UNCONSTRAINED :
					instanceLinkIterator = instanceLinks.iterator();
					break;
				case FIRST :
					instanceLinkIterator = instanceLinks.subList(0, 1).iterator(); 
					break;
				case LAST :
					instanceLinkIterator = instanceLinks.subList(instanceLinks.size() - 1, instanceLinks.size()).iterator(); 
					break;
				default :
					instanceLinkIterator = instanceLinks.iterator();
					break;
			}
		}
		
		return new MLSDMTraverseLinkMatchState(instanceLinkIterator);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		EObject instanceObject = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), instanceObject, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		boolean valid = false;
		MLSDMTraverseLinkMatchState ms = (MLSDMTraverseLinkMatchState) matchState;
		Iterator<? extends Object> featureIterator = ms.getFeatureIterator();
		Object targetInstance = null;
		while(!valid && featureIterator.hasNext()) {
			targetInstance = featureIterator.next();
			valid = matcher.instanceMatchesSPO(targetInstance, targetNode.getSpo()) && matcher.checkMatchingHistory(this.getPatternConstraint(), targetInstance);
		}
		
		transaction.setValid(valid);
		
		if(valid) {
			transaction.setTargetInstance(targetInstance);
		}
		else {
			matcher.getNotificationEmitter().storyPatternObjectNotBound(targetNode.getSpo(), matcher.getVariablesScope(),
					matcher);
		}
		

		matcher.updateMatchingHistory(this.getPatternConstraint(), targetInstance);
		
		return transaction;
	}

	public EStructuralFeature getFeature() {
		return feature;
	}
	
	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getSourceNode() {
		return sourceNode;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getTargetNode() {
		return targetNode;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse\t(" + sourceNode.getSpo().getName() + " -" + feature.getName() + "-> " + targetNode.getSpo().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMLinkTraverseMatchingAction(id, matcher, clonePatternConstraint, link, feature, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}
}
