package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Map;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check of a constraint based on a map entry link in the
 * story pattern. It can be executed once mappings for the
 * source and value as well as the key node of the link
 * have been created.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMMapCheckMatchingAction extends MLSDMCheckMatchingAction{

	/*
	 * The map entry link.
	 */
	private MapEntryLink link;
	
	/*
	 * The source node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * The value node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode;
	
	/*
	 * The key node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode;
	
	public MLSDMMapCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, MapEntryLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode ) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.valueNode = valueNode;
		this.keyNode = keyNode;
		addRequiredBinding(sourceNode);
		addRequiredBinding(valueNode);
		addRequiredBinding(keyNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		if(!sourceNode.isBound() || (valueNode != null && !valueNode.isBound()) || !keyNode.isBound()) {
			return null;
		}
		else{
			return new MLSDMCheckOnceMatchState();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!sourceNode.isBound() || (valueNode != null && !valueNode.isBound()) || !keyNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return PatternConstraint.CHECK_ONLY;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkEntry());
		}
		return transaction;
	}

	/**
	 * Check whether the map contains an entry matching the
	 * mappings for key and value node.
	 * 
	 * @return	True if a matching map entry exists.
	 */
	@SuppressWarnings("unchecked")
	private boolean checkEntry() {
		EObject sourceInstance = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		Object keyInstance = matcher.getVariable(keyNode.getSpo().getName()).getValue();
		
		boolean success;
		
		if(valueNode != null) {
			Object valueInstance = matcher.getVariable(valueNode.getSpo().getName()).getValue();
			Object entryValue = ((EMap<Object, Object>) sourceInstance.eGet(link.getFeature())).get(keyInstance);
			success = valueInstance == entryValue;
		}
		else {
			success = ((EMap<Object, Object>) sourceInstance.eGet(link.getFeature())).containsKey(keyInstance);
		}
		
		if(success) {
			matcher.getNotificationEmitter().linkCheckSuccessful(link.getSource(), sourceInstance, link, link.getTarget(), keyInstance, matcher.getVariablesScope(), matcher);
		}
		else {
			matcher.getNotificationEmitter().linkCheckFailed(link.getSource(), sourceInstance, link, link.getTarget(), keyInstance, matcher.getVariablesScope(), matcher);
		}
		
		return success;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkEntry();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check \t\t (" + link.getSource().getName() + " -" + link.getFeature().getName() + "-> " + "(" + link.getTarget().getName() + ", " + link.getValueTarget().getName() + "))";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMMapCheckMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(valueNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(keyNode));
	}
	
}
