package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternConstraint;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternNode;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMSearchModel;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.ContainmentLink;
import de.mdelab.mlstorypatterns.ExpressionLink;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.MatchTypeEnum;
import de.mdelab.mlstorypatterns.MlstorypatternsPackage;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check of a negative application condition formulated
 * over the story pattern. It can be executed once all
 * dependencies of the negative application condition
 * in the parent story pattern have been mapped.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMNACCheckMatchingAction extends MLSDMCheckMatchingAction {

	/*
	 * Pattern of the negative application condition.
	 */
	private StoryPattern nacPattern;
	
	private MLSDMSearchModelBasedNACMatcher nacMatcher;
	
	public MLSDMNACCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, StoryPattern nacPattern,
			Collection<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requiredPatternNodes) {
		super(id, matcher);
		this.nacPattern = nacPattern;
		this.nacMatcher = createNACMatcher();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> requiredPatternNode:requiredPatternNodes) {
			addRequiredBinding(requiredPatternNode);
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:getRequiredBindings()) {
			if(!patternNode.isBound()) {
				return null;
			}
		}
		return new MLSDMCheckOnceMatchState();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:getRequiredBindings()) {
			if(!patternNode.isBound()) {
				return PatternConstraint.MATCHING_NOT_POSSIBLE;
			}
		}
		return PatternConstraint.ALWAYS_MATCH_LAST; //TODO analyse the nac pattern to give a better estimate
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkNAC());
		}
		return transaction;
	}

	/**
	 * Perform the actual NAC check.
	 * 
	 * @return	True if the current match can not be extended
	 *			with elements matching the NAC pattern.
	 */
	private boolean checkNAC() {
		try {
			nacMatcher.rollBack();
			nacMatcher.setVariablesScope(matcher.getVariablesScope());
			nacMatcher.ensureConsistentSearchModel();
			return !nacMatcher.findNextMatch();
		} catch (SDMException e) {
			// TODO throw this exception
			e.printStackTrace();
		}
		return false;
	}
	
	private MLSDMSearchModelBasedNACMatcher createNACMatcher() {
		try {
			MLSDMNACSearchModelBuilder nacSearchModelBuilder = new MLSDMNACSearchModelBuilder() {
				
				@Override
				protected Map<AbstractStoryPatternObject, MLSDMPatternNode> createPatternNodes(MLSDMSearchModel searchModel, StoryPattern storyPattern) {
					Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes = super.createPatternNodes(searchModel, storyPattern);
					
					/*
					 * Also account for nodes outside but connected to
					 * the nac pattern.
					 */
					outsideNodes = new HashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
					int id = patternNodes.size();
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:getRequiredBindings()) {
						MLSDMPatternNode outsideNode = new MLSDMPatternNode(id++, searchModel, patternNode.getSpo());
						outsideNode.setBound(true);
						patternNodes.put(patternNode.getSpo(), outsideNode);
						outsideNodes.add(outsideNode);
					}
					return patternNodes;
				}
				
				@Override
				protected Collection<MLSDMPatternConstraint> createLinkPatternConstraints(StoryPattern storyPattern, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
					List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
	
					Set<AbstractStoryPatternLink> consideredLinks = new HashSet<AbstractStoryPatternLink>();
					consideredLinks.addAll(storyPattern.getStoryPatternLinks());
					for(AbstractStoryPatternLink link:storyPattern.getContainingPattern().getStoryPatternLinks()) {
						if(link.getSource().getStoryPattern() == storyPattern || link.getTarget().getStoryPattern() == storyPattern) {
							consideredLinks.add(link);
						}
					}
					
					for(AbstractStoryPatternLink link:consideredLinks) {
						/*
						 * Only consider links that are not part of a NAC
						 * and that are not being created.
						 */
						if((link.getModifier() != ModifierEnum.CREATE || link.getMatchType() == MatchTypeEnum.OPTIONAL) && 
								(link.getSource().getStoryPattern() == storyPattern || link.getSource().getStoryPattern() == storyPattern.getContainingPattern()) &&
								(link.getTarget().getStoryPattern() == storyPattern || link.getTarget().getStoryPattern() == storyPattern.getContainingPattern())) {
							final EClass linkEClass = link.eClass();
							
							MLSDMPatternConstraint patternConstraint = null;
							if (linkEClass == MlstorypatternsPackage.Literals.STORY_PATTERN_LINK) {
								patternConstraint = createLinkPatternConstraint((StoryPatternLink) link, patternNodes);
							}
							else if(linkEClass == MlstorypatternsPackage.Literals.CONTAINMENT_LINK) {
								patternConstraint = createContainmentLinkPatternConstraint((ContainmentLink) link, patternNodes);
							}
							else if(linkEClass == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK) {
								patternConstraint = createMapEntryLinkPatternConstraint((MapEntryLink) link, patternNodes);
							}
							else if(linkEClass == MlstorypatternsPackage.Literals.EXPRESSION_LINK) {
								patternConstraint = createExpressionLinkPatternConstraint((ExpressionLink) link, patternNodes);
							}
							
							if(patternConstraint != null){
								patternConstraint.setOptional(link.getMatchType() == MatchTypeEnum.OPTIONAL);
								patternConstraints.add(patternConstraint);
							}
						}
					}
					return patternConstraints;
				}
			};
			
			return new MLSDMSearchModelBasedNACMatcher(nacPattern, matcher.getVariablesScope(), matcher.getFacadeFactory(), matcher.getExpressionInterpreterManager(), matcher.getExpressionAnalyzerManager(), matcher.getStrategyFactory(), matcher.getReferenceAdapter(), matcher.getNotificationEmitter(), matcher.getInterpreterParameters(), nacSearchModelBuilder);
		} catch (SDMException e) {
			// TODO throw this exception
			e.printStackTrace();
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkNAC();
	}

	@Override
	public void reset() {
		nacMatcher.reset();
		super.reset();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check\t\t(NAC: " + nacPattern.getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requiredClones
				= new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> requirement:getRequiredBindings()) {
			requiredClones.add((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(requirement));
		}
		return new MLSDMNACCheckMatchingAction(id, nacMatcher, nacPattern, requiredClones);
	}
	
}
