package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.expressions.interpreter.core.ExpressionsInterpreterException;
import de.mdelab.expressions.interpreter.core.ExpressionInterpreterManager;
import de.mdelab.expressions.interpreter.core.variables.Variable;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMReferenceIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMStrategyFactory;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * This is a slightly adapted search model based matcher for
 * matching patterns of negative application conditions. It
 * takes into account that there are nodes outside the negative
 * pattern for which mappings have been created by a parent
 * matcher.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMSearchModelBasedNACMatcher extends MLSDMSearchModelBasedMatcher {

	/*
	 * Collection of pattern nodes of the parent pattern which
	 * are connected to nodes of the negative pattern.
	 */
	protected Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> outsideNodes;
	
	public MLSDMSearchModelBasedNACMatcher(StoryPattern storyPattern,
			NotifierVariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> variablesScope,
			MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> facadeFactory,
			ExpressionInterpreterManager<EClassifier, EStructuralFeature, MLExpression> expressionInterpreterManager,
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> expressionAnalyzerManager,
			MLSDMStrategyFactory strategyFactory,
			MLSDMReferenceIndex referenceAdapter,
			NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
			Map<String, Object> interpreterParameters,
			MLSDMNACSearchModelBuilder nacSearchModelBuilder)
			throws SDMException {
		super(storyPattern, variablesScope, facadeFactory, expressionInterpreterManager, expressionAnalyzerManager,
				strategyFactory, referenceAdapter, notificationEmitter, interpreterParameters, nacSearchModelBuilder);
	}	

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBasedMatcher#analyzeStoryPatternObjects()
	 */
	protected boolean analyzeStoryPatternObjects()
	throws SDMException {
		/*
		 * analyze all story pattern objects
		 */
		for (final PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode : searchModel.getPatternNodes())
		{
			if(patternNode.isBound()) {
				continue;
			}
			
			AbstractStoryPatternObject spo = patternNode.getSpo();
			
			boolean bound = this.spoFacade.isBound(spo);

			if (!bound && this.spoFacade.isMaybeBound(spo))
			{
				/*
				 * Check if a variable exists with that name
				 */
				bound = this.getVariablesScope().variableExists(this.spoFacade.getName(spo));
			}
			
			/*
			 * Treat story pattern objects outside but connected to
			 * the nac pattern as bound from the start.
			 */
			if (bound || outsideNodes.contains(patternNode))
			{
				/*
				 * Check if there is an assignment expression
				 */
				final MLExpression assignmentExpression = this.spoFacade.getAssignmentExpression(spo);

				Variable<EClassifier> variable = null;

				if (assignmentExpression != null)
				{
					/*
					 * Evaluate the expression
					 */
					Variable<EClassifier> result;
					
					try {
						result = this.getExpressionInterpreterManager().evaluateExpression(
								this.spoFacade.getAssignmentExpression(spo), null, null, this.getVariablesScope());
					} catch ( final ExpressionsInterpreterException ex ) {
						throw new SDMException( ex );
					}

					if (result != null)
					{
						/*
						 * Create a new variable with the name and classifier of
						 * the story pattern object
						 */
						variable = this.getVariablesScope().createVariable(this.spoFacade.getName(spo), this.spoFacade.getClassifier(spo),
								result.getValue());
					}
					else
					{
						throw new SDMException("The expression '" + this.spoFacade.getAssignmentExpression(spo)
								+ "' could not be evaluated.");
					}
				}
				else
				{
					variable = this.getVariablesScope().getVariable(this.spoFacade.getName(spo));

					if (variable == null)
					{
						/*
						 * There is no variable with that name.
						 */
						this.getNotificationEmitter().storyPatternObjectNotBound(spo, this.getVariablesScope(), this);

						return false;
					}
				}

				/*
				 * Check that the type of the existing variable matches the type
				 * of the story pattern object, check isomorphism condition and
				 * object constraints
				 */
				if ((variable.getValue() != null) && this.checkTypeConstraint(variable.getValue(), spo)
						&& this.checkIsomorphism(variable.getValue()))
				{
					MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
							new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(boundSPOConstraint, this);
					transaction.setSPO(spo);
					transaction.setTargetInstance(variable.getValue());
					transaction.setValid(true);
					transaction.addPatternNode(patternNode);
					
					transaction.commit();
					matchingStack.push(transaction);

					selectionStrategy.update(transaction.getAffectedPatternNodes(), true);
					
					this.getNotificationEmitter().storyPatternObjectBound(spo, variable.getValue(), this.getVariablesScope(), this);
				}
				else
				{
					/*
					 * Constraints are not satisfied or instance object already
					 * bound.
					 */
					this.getNotificationEmitter().storyPatternObjectNotBound(spo, this.getVariablesScope(), this);

					return false;
				}
			}
			else
			{
				/*
				 * Delete existing variables of unbound story pattern objects
				 */
				this.getVariablesScope().deleteVariable(this.spoFacade.getName(spo));
			}
		}

		return true;
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.StoryPatternMatcher#findNextMatch()
	 */
	@Override
	public boolean findNextMatch() throws SDMException {
		if(searchModel == null) {
			this.getNotificationEmitter().storyPatternInitializationStarted(this.getStoryPattern(), this.getVariablesScope(), this);

			this.searchModel = searchModelBuilder.createSearchModel(getStoryPattern());
			for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:searchModel.getPatternConstraints()) {
				matchingHistory.put(pc, new HashSet<Object>());
			}
			
			this.outsideNodes = ((MLSDMNACSearchModelBuilder)searchModelBuilder).outsideNodes;

			selectionStrategy.initializeWithModel(searchModel);
			
			this.getNotificationEmitter().storyPatternInitializationFinished(this.getStoryPattern(), this.getVariablesScope(), this);
		}
		
		return super.findNextMatch();
	}

	public void rollBack() {
		while(matchingStack.peek() != null) {
			rollBackLastTransaction(true);
		}
	}
	
	@Override
	public void reset() {
		super.reset();
		
		/*
		 * outside nodes should always remain bound
		 */
		if(searchModel != null) {
			for(PatternNode<?, ?, ?, ?, ?, ?> outsideNode:outsideNodes) {
				outsideNode.setBound(true);
			}
		}
	}

	public void ensureConsistentSearchModel() {
		if(searchModel != null) {
			selectionStrategy.initializeWithModel(searchModel);
		}
	}

}
