package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLExpression;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLStringExpression;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.icl.iCL.ICLVariableAttribute;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.FeatureAccess;
import de.mdelab.sdm.interpreter.icl.ICLExpressionInterpreter;

/**
 * This is an expression analyzer implementation for the
 * Index Constraint Language (ICL).
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMICLExpressionAnalyzer extends MLSDMExpressionAnalyzer { 
	
	/*
	 * ICL interpreter.
	 */
	private ICLExpressionInterpreter<MLExpression> interpreter;
	
	public MLSDMICLExpressionAnalyzer() {
		this(null);	//TODO factory
	}
	
	public MLSDMICLExpressionAnalyzer(ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> analyzerManager) {
		super(analyzerManager);
		interpreter = new ICLExpressionInterpreter<MLExpression>();
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#getVariableNames(java.lang.Object)
	 */
	@Override
	public Collection<String> getVariableNames(MLExpression expression) {
		Object ast = parseAST(expression);
		Set<String> names = new HashSet<String>();
		if(ast != null && ((EObject) ast).eClass() == ICLPackage.Literals.ICL_CONSTRAINT) {
			for(ICLParameter parameter:((ICLConstraint) ast).getParameters().getList()) {
				if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
					Collection<String> variableNamesMin = getVariableNames(((ICLRange) parameter).getMin());
					if(variableNamesMin != null) {
						names.addAll(variableNamesMin);
					}
					
					Collection<String> variableNamesMax = getVariableNames(((ICLRange) parameter).getMax());
					if(variableNamesMax != null) {
						names.addAll(variableNamesMax);
					}
				}
				else {
					Collection<String> variableNames = getVariableNames((ICLValue) parameter);
					if(variableNames != null) {
						names.addAll(variableNames);
					}
				}
			}
		}
		return names;
	}

	/**
	 * Retrieve the names of all variables referenced in an
	 * ICL value.
	 * 
	 * @param value
	 * 			The ICL value.
	 * @return	A collection containing the names of all
	 * 			referenced variables.
	 */
	private Collection<String> getVariableNames(ICLValue value) {
		if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE) {
			return Collections.singleton(((ICLVariable) value).getName());
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE_ATTRIBUTE) {
			return Collections.singleton(((ICLVariableAttribute) value).getVariable().getName());
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_STRING_EXPRESSION) {
			return analyzerManager.getVariableNames(((ICLStringExpression) value).getExpression());
		}
		else{
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#getVariableNames(java.lang.Object)
	 */
	@Override
	public Collection<FeatureAccess<EStructuralFeature>> getFeatureAccesses(MLExpression expression, EClassifier contextClassifier) {
		Object ast = parseAST(expression);
		Collection<FeatureAccess<EStructuralFeature>> accesses = new HashSet<FeatureAccess<EStructuralFeature>>();
		if(ast != null && ((EObject) ast).eClass() == ICLPackage.Literals.ICL_CONSTRAINT) {
			for(ICLParameter parameter:((ICLConstraint) ast).getParameters().getList()) {
				if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
					Collection<FeatureAccess<EStructuralFeature>> featureAccessesMin = getFeatureAccesses(((ICLRange) parameter).getMin());
					if(featureAccessesMin != null) {
						accesses.addAll(featureAccessesMin);
					}
					
					Collection<FeatureAccess<EStructuralFeature>> featureAccessesMax = getFeatureAccesses(((ICLRange) parameter).getMax());
					if(featureAccessesMax != null) {
						accesses.addAll(featureAccessesMax);
					}
				}
				else {
					Collection<FeatureAccess<EStructuralFeature>> featureAccesses = getFeatureAccesses((ICLValue) parameter);
					if(featureAccesses != null) {
						accesses.addAll(featureAccesses);
					}
				}
			}
		}
		return accesses;
	}
	
	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccesses(
			ICLValue value) {
		if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE) {
			return Collections.emptySet();
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE_ATTRIBUTE) {
			ICLVariableAttribute variableAttribute = (ICLVariableAttribute) value;
			String name = variableAttribute.getVariable().getName();
			return Collections.singleton(new FeatureAccess<EStructuralFeature>(name, getVariableFeature(name, variableAttribute.getAttribute())));
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_STRING_EXPRESSION) {
			return analyzerManager.getFeatureAccesses(((ICLStringExpression) value).getExpression());
		}
		else{
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#parseAST(java.lang.Object)
	 */
	@Override
	public Object parseAST(MLExpression expression, EClassifier contextClassifier) {
		return parseAST(expression);
	}
	
	protected Object parseAST(MLExpression expression) {
		try {
			return interpreter.parseExpression(((MLStringExpression) expression).getExpressionString());
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Convert an ICL expression AST into a textual representation.
	 * 
	 * @param expression
	 * 			The AST of the expression.
	 * @return	A textual representation of the expression.
	 */
	public String getExpressionString(ICLExpression expression) {
		return interpreter.getExpressionString(expression);
	}

}
