package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy;

/**
 * This is a pattern constraint selection strategy based on a cost
 * model for pattern constraints. In each step, the cheapest pattern
 * constraint to handle according to the cost mode is selected for
 * checking.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMCheapestFirstSelectionStrategy extends OrderProducingSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {
	
	private static boolean RESET_COST_STRATEGY = false;
	
	/*
	 * Strategy implementing the cost model.
	 */
	protected MLSDMMatchingCostStrategy costStrategy;

	/*
	 * Available checks
	 */
	protected List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> pendingChecks;
	
	public MLSDMCheapestFirstSelectionStrategy() {
		this.pendingChecks = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
	}
	
	protected MLSDMMatchingCostStrategy createMatchingCostStrategy() {
		return new MLSDMCardinalityMatchingCostStrategy();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#selectNextConstraint()
	 */
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> selectNextConstraint() {
		if(!pendingChecks.isEmpty()) {
			return pendingChecks.remove(pendingChecks.size() - 1);
		}
		
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> cheapestConstraint = null;
		double minCost = PatternConstraint.MATCHING_NOT_POSSIBLE;
		
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> cheapestConstraintOptional = null;
		double minCostOptional = PatternConstraint.MATCHING_NOT_POSSIBLE;
		
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> constraint:searchModel.getPatternConstraints()) {
			if(constraint.isActive()) {
				continue;
			}
			
			double cost = costStrategy.calculateMatchingCost(constraint);
			if(cost == PatternConstraint.CHECK_ONLY) {
				pendingChecks.add(constraint);
			}
			
			if(cost == PatternConstraint.MATCHING_NOT_POSSIBLE) {
				continue;
			}
			
			if(!constraint.isOptional()) {
				if(cost < minCost) {
					cheapestConstraint = constraint;
					minCost = cost;
				}
			}
			else {
				if(cost < minCostOptional) {
					cheapestConstraintOptional = constraint;
					minCostOptional = cost;
				}
			}
		}
		
		if(!pendingChecks.isEmpty()) {
			return pendingChecks.remove(pendingChecks.size() - 1);
		}
		else if(cheapestConstraint != null) {
			return cheapestConstraint;
		}
		else {
			return cheapestConstraintOptional;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.PatternConstraintSelectionStrategy#doInitialize()
	 */
	@Override
	protected void doInitialize() {
		if(costStrategy == null || RESET_COST_STRATEGY) {
			costStrategy = createMatchingCostStrategy();
			costStrategy.initialize();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.PatternConstraintSelectionStrategy#doRollBackLastPop()
	 */
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> doRollBackLastPop() {
		pendingChecks.clear();
		return super.doRollBackLastPop();
	}
	
	@Override
	public void reset() {
		super.reset();
		pendingChecks.clear();
	}

}
