package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexStagedMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexStepMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkContainmentReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.mlindices.StagedIndex;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.NotifierVariablesScope;

/**
 * Table calculator for hybrid and static matching strategies.
 * This computes a table containing an estimate for the size of
 * the search space remaining for each search model configuration.
 * Different cost models for computing a static cost estimate for
 * each individual matching action can be used.
 * Note that the computation takes exponential time in the size
 * of the query graph.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMContinuationCostCalculator {

	/**
	 * Enumeration of possible cost models for cost estimates of
	 * individual matching actions.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	public static enum CostModel {AVERAGE_CASE, WORST_CASE, BEST_CASE};
	
	/**
	 * This is an entry in the computed static cost table.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	public class Adornment {

		/**
		 * Search model configuration in this entry
		 */
		public boolean[] bindings;
		
		/**
		 * Collection of pattern constraints that have not yet
		 * been handled in this entry's configuration.
		 */
		public Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> unhandledConstraints;
		
		/**
		 * Size of the search space remaining in this entry's
		 * configuration.
		 */
		public double totalCost;
		
		/**
		 * Best computed matching order for the remainder of the
		 * query in this entry's configuration (pattern constraints).
		 */
		public List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder;
		
		/**
		 * Best computed matching order for the remainder of the
		 * query in this entry's configuration (matching actions).
		 */
		public List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions;
		
		/**
		 * Total filtering effect of all pattern constraints that
		 * have already been checked in this entry's configuration.
		 */
		public double appliedFilters;
		
		/**
		 * Size of the search front after reaching this entry's
		 * configuration. Only required for fully static strategy.
		 */
		protected double front = 1;
		
		public Adornment() {
			this.bindings = new boolean[searchModel.getPatternNodes().size()];
			for(int i = 0; i < bindings.length; i++) {
				this.bindings[i] = true;
			}
			
			this.unhandledConstraints = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(searchModel.getPatternConstraints());
			this.totalCost = 0;
			this.matchingOrder = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			this.actions = new ArrayList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			this.appliedFilters = 1;
		}

		public Adornment(	boolean[] bindings,
							Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> checkedConstraints,
							double remainingCost,
							List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder,
							List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions,
							double appliedFilters) {
			this(bindings, checkedConstraints, remainingCost, matchingOrder, actions, appliedFilters, 1.0);
		}
		
		public Adornment(	boolean[] bindings,
				Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> checkedConstraints,
				double remainingCost,
				List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder,
				List<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions,
				double appliedFilters,
				double front) {
			this.bindings = bindings;
			this.unhandledConstraints = checkedConstraints;
			this.totalCost = remainingCost;
			this.matchingOrder = matchingOrder;
			this.actions = actions;
			this.appliedFilters = appliedFilters;
			this.front = front;
		}
		
		@Override
		public String toString() {
			StringBuffer sb = new StringBuffer();
			sb.append(totalCost + "\t-\t[");
			boolean separate = false;
			for(int j = 0; j < bindings.length; j++) {
				if(!bindings[j]) {
					continue;
				}
				if(separate) {
					sb.append(", ");
				}
				sb.append(((AbstractStoryPatternObject) searchModel.getPatternNodes().get(j).getSpo()).getName());
				separate = true;
			}
			sb.append("]");
			return sb.toString();
		}
		
	}
	
	protected SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel;
	protected MLSDMMetadataIndex metadataAdapter;
	protected CostModel costModel;
	protected NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope;
	protected Map<String, PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> nameMappings;
	
	public MLSDMContinuationCostCalculator(SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			MLSDMMetadataIndex metadataAdapter, CostModel costModel, NotifierVariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		this.searchModel = searchModel;
		this.metadataAdapter = metadataAdapter;
		this.costModel = costModel;
		this.variablesScope = variablesScope;
		
		this.nameMappings = new HashMap<String, PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:this.searchModel.getPatternNodes()) {
			nameMappings.put(node.getSpo().getName(), node);
		}
	}
	
	public Adornment[] calculateCotinuationCosts() {
		Adornment[] table = new Adornment[(int) Math.pow(2, searchModel.getPatternNodes().size())];
		Adornment initial = new Adornment();
		table[(int) Math.pow(2, searchModel.getPatternNodes().size()) - 1] = initial;

		@SuppressWarnings("unchecked")
		Collection<Adornment>[] configurations = (Collection<Adornment>[]) new Collection[searchModel.getPatternNodes().size() + 1];
		for(int i = 0; i < configurations.length; i++) {
			configurations[i] = new LinkedHashSet<Adornment>();
		}
		configurations[configurations.length - 1].add(initial);
		
		for(int i = configurations.length - 1; i >= 0; i--) {
			Collection<Adornment> adornments = configurations[i];
			for(Adornment a:adornments) {
				Collection<Adornment> predecessors = computePredecessors(a);
				for(Adornment predecessor:predecessors) {
					int bindingCode = getBindingCode(predecessor);
					if(table[bindingCode] == null || table[bindingCode].totalCost > predecessor.totalCost || (table[bindingCode].totalCost == predecessor.totalCost && predecessor.matchingOrder.toString().compareTo(table[bindingCode].matchingOrder.toString()) > 0)) {
						int boundNodeNumber = getBoundNodeNumber(predecessor);
						configurations[boundNodeNumber].remove(table[bindingCode]);
						configurations[boundNodeNumber].add(predecessor);
						table[bindingCode] = predecessor;
					}
				}
			}
		}
		return table;
	}

	protected int getBoundNodeNumber(Adornment adornment) {
		int i = 0;
		for(boolean b:adornment.bindings) {
			if(b) {
				i++;
			}
		}
		return i;
	}

	protected int getBindingCode(Adornment adornment) {
		int code = 0;
		for(int i  = 0; i < adornment.bindings.length; i++) {
			if(adornment.bindings[i]) {
				code = code | (1 << i);
			}
		}
		return code;
	}

	protected Collection<Adornment> computePredecessors(Adornment a) {
		Collection<Adornment> predecessors = new ArrayList<Adornment>();
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:a.unhandledConstraints) {
			for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
				/*
				 * skip explicit checks
				 */
				if(ma.getRequiredBindings().size() == ma.getPatternConstraint().getDependencies().size()) {
					continue;
				}
				
				if(ma instanceof MLSDMIndexStagedMatchingAction) {
					predecessors.addAll(getIndexStagedPredecessors(a, (MLSDMIndexStagedMatchingAction) ma));
				}
				else {
					Adornment predecessor = getPredecessor(a, ma);
					if(predecessor != null) {
						predecessors.add(predecessor);
					}
				}
			}
		}
		return predecessors;
	}

	protected Collection<Adornment> getIndexStagedPredecessors(
			Adornment a, MLSDMIndexStagedMatchingAction ma) {
		Collection<Adornment> predecessors = new ArrayList<Adornment>();
		
		ICLConstraint constraint = ((ICLConstraint) ma.getPatternConstraint().getConstraint());
		StagedIndex index = (StagedIndex) variablesScope.getVariable(constraint.getIndexID()).getValue();
		int parameterNumber = constraint.getParameters().getList().size();
		boolean[] currentBindings = new boolean[searchModel.getPatternNodes().size()];
		System.arraycopy(a.bindings, 0, currentBindings, 0, currentBindings.length);
		
		for(int fromLevel = parameterNumber - 1; fromLevel >= 0; fromLevel--) {
			if(!(constraint.getParameters().getList().get(fromLevel) instanceof ICLVariable)) {
				return new ArrayList<Adornment>();
			}
			ICLVariable parameter = (ICLVariable) constraint.getParameters().getList().get(fromLevel);
			
			double cost = estimateStagedIndexCost(index, fromLevel);			
			boolean[] derivedBindings = new boolean[searchModel.getPatternNodes().size()];
			System.arraycopy(currentBindings, 0, derivedBindings, 0, derivedBindings.length);
			PatternNode<?, AbstractStoryPatternObject, ?, ?, ?, ?> node = nameMappings.get(((ICLVariable) parameter).getName());
			derivedBindings[node.getId()] = false;
			
			predecessors.add(createDerivedAdornment(a, derivedBindings, cost, parameterNumber - fromLevel, ma));
			currentBindings = derivedBindings;
		}
		
		return predecessors;
	}

	protected double estimateStagedIndexCost(StagedIndex index, int fromLevel) {
		switch(costModel) {
			case AVERAGE_CASE :
				long levelSize = index.getSize(fromLevel);
				return index.size() / (levelSize > 0 ? levelSize : 1);
			case WORST_CASE :
				return index.getMaxChildren(fromLevel);
			case BEST_CASE :
				return index.getMinChildren(fromLevel);
			default :
				levelSize = fromLevel > 0 ? index.getSize(fromLevel - 1) : 1;
				return index.size() / (levelSize > 0 ? levelSize : 1);
		}
	}

	protected Adornment createDerivedAdornment(Adornment a, boolean[] derivedBindings,
			double cost, int newBindings, MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> checkedConstraints = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.unhandledConstraints);
		LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder = new LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.matchingOrder);
		LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions = new LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.actions);
		for(Iterator<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> it = checkedConstraints.iterator(); it.hasNext();) {
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc2 = it.next();
			if(!isChecked(pc2, derivedBindings)) {
				it.remove();
				if(pc2 != ma.getPatternConstraint()) {
					matchingOrder.add(0, pc2);
					actions.add(0, pc2.getExplicitCheckAction());
				}
			}
		}
		matchingOrder.add(0, ma.getPatternConstraint());
		actions.add(0, ma);
		
		return new Adornment(derivedBindings, checkedConstraints, (cost * newBindings) + (cost * a.totalCost), matchingOrder, actions, 1);
	}

	protected Adornment getPredecessor(Adornment a,
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		double cost = estimateCost(ma);
		
		if(cost != PatternConstraint.MATCHING_NOT_POSSIBLE) {
			boolean[] derivedBindings = new boolean[searchModel.getPatternNodes().size()];
			System.arraycopy(a.bindings, 0, derivedBindings, 0, derivedBindings.length);
			int newBindings = apply(ma, derivedBindings);

			return createDerivedAdornment(a, derivedBindings, cost, newBindings, ma);
		}
		
		return null;
	}

	protected int apply(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma, boolean[] bindings) {
		int newBindings = 0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:ma.getPatternConstraint().getDependencies()) {
			if(ma.getRequiredBindings().contains(pn)) {
				continue;
			}
			int index = pn.getId();
			bindings[index] = false;
			newBindings++;
		}
		return newBindings;
	}

	protected boolean isChecked(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc2, boolean[] bindings) {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:pc2.getDependencies()) {
			int index = pn.getId();
			if(!bindings[index]) {
				return false;
			}
		}
		return true;
	}

	public double estimateCost(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		switch(costModel) {
			case AVERAGE_CASE :
				return estimateAverageCost(ma);
			case WORST_CASE :
				return estimateWorstCost(ma);
			case BEST_CASE :
				return estimateBestCost(ma);
			default :
				return estimateAverageCost(ma);
		}
	}
	
	protected double estimateAverageCost(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMDomainMatchMatchingAction) {
			return (double) metadataAdapter.getDomain((EClass) ma.getPatternConstraint().getConstraint()).size();
		}
		else if(ma instanceof MLSDMLinkTraverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseMatchingAction) ma).getFeature();
			EClass sourceClass = (EClass) ((MLSDMLinkTraverseMatchingAction) ma).getSourceNode().getSpo().getType();
			double referenceNumber = (double) metadataAdapter.getReferences(reference).size();
			double domainSize = (double) metadataAdapter.getDomain(sourceClass).size();
			return 	domainSize != 0 ? referenceNumber / domainSize : 0;
		}
		else if(ma instanceof MLSDMLinkTraverseReverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseReverseMatchingAction) ma).getFeature();
			EClass targetClass = (EClass) ((MLSDMLinkTraverseReverseMatchingAction) ma).getTargetNode().getSpo().getType();
			double referenceNumber = (double) metadataAdapter.getReferences(reference).size();
			double domainSize = (double) metadataAdapter.getDomain(targetClass).size();
			return domainSize != 0 ? referenceNumber / domainSize : 0;
		}
		else if(ma instanceof MLSDMLinkContainmentReverseMatchingAction) {
			return 	1;
		}
		else if(ma instanceof MLSDMLinkLookupMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkLookupMatchingAction) ma).getFeature();
			return 	metadataAdapter.getReferences(reference).size();
		}
		else if(ma instanceof MLSDMIndexStepMatchingAction) {
			ICLConstraint constraint = (ICLConstraint) ma.getPatternConstraint().getConstraint();
			StagedIndex index = (StagedIndex) variablesScope.getVariable(constraint.getIndexID()).getValue();
			double cost = 1;
			for(int fromLevel = 0; fromLevel < constraint.getParameters().getList().size(); fromLevel ++) {
				ICLParameter parameter = constraint.getParameters().getList().get(fromLevel);
				if(parameter instanceof ICLVariable &&
						((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter) != null &&
						!ma.getRequiredBindings().contains(((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter)))  {
					double nextSize = index.getSize(fromLevel + 1);
					double fromSize = index.getSize(fromLevel);
					cost *=  fromSize != 0 ? nextSize / fromSize : 0;
				}
			}
			return cost;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	protected double estimateWorstCost(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMDomainMatchMatchingAction) {
			return (double) metadataAdapter.getDomain((EClass) ma.getPatternConstraint().getConstraint()).size();
		}
		else if(ma instanceof MLSDMLinkTraverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseMatchingAction) ma).getFeature();
			return 	(double) metadataAdapter.getMaxCardinality(reference);
		}
		else if(ma instanceof MLSDMLinkTraverseReverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseReverseMatchingAction) ma).getFeature();
			return 	(double) metadataAdapter.getMaxInverseCardinality(reference);
		}
		else if(ma instanceof MLSDMLinkContainmentReverseMatchingAction) {
			return 	1;
		}
		else if(ma instanceof MLSDMLinkLookupMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkLookupMatchingAction) ma).getFeature();
			return 	metadataAdapter.getReferences(reference).size();
		}
		else if(ma instanceof MLSDMIndexStepMatchingAction) {
			ICLConstraint constraint = (ICLConstraint) ma.getPatternConstraint().getConstraint();
			StagedIndex index = (StagedIndex) variablesScope.getVariable(constraint.getIndexID()).getValue();
			double cost = 1;
			for(int fromLevel = 0; fromLevel < constraint.getParameters().getList().size(); fromLevel ++) {
				ICLParameter parameter = constraint.getParameters().getList().get(fromLevel);
				if(parameter instanceof ICLVariable &&
						((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter) != null &&
						!ma.getRequiredBindings().contains(((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter)))  {
					cost *= index.getMaxChildren(fromLevel);
				}
			}
			return cost;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}
	
	protected double estimateBestCost(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMDomainMatchMatchingAction) {
			return (double) metadataAdapter.getDomain((EClass) ma.getPatternConstraint().getConstraint()).size();
		}
		else if(ma instanceof MLSDMLinkTraverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseMatchingAction) ma).getFeature();
			return 	(double) metadataAdapter.getMinCardinality(reference);
		}
		else if(ma instanceof MLSDMLinkTraverseReverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseReverseMatchingAction) ma).getFeature();
			return 	(double) metadataAdapter.getMinInverseCardinality(reference);
		}
		else if(ma instanceof MLSDMLinkLookupMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkLookupMatchingAction) ma).getFeature();
			return 	metadataAdapter.getReferences(reference).size();
		}
		else if(ma instanceof MLSDMIndexStepMatchingAction) {
			ICLConstraint constraint = (ICLConstraint) ma.getPatternConstraint().getConstraint();
			StagedIndex index = (StagedIndex) variablesScope.getVariable(constraint.getIndexID()).getValue();
			double cost = 1;
			for(int fromLevel = 0; fromLevel < constraint.getParameters().getList().size(); fromLevel ++) {
				ICLParameter parameter = constraint.getParameters().getList().get(fromLevel);
				if(parameter instanceof ICLVariable &&
						((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter) != null &&
						!ma.getRequiredBindings().contains(((MLSDMIndexStepMatchingAction) ma).getPatternNode((ICLVariable) parameter)))  {
					cost *= index.getMinChildren(fromLevel);
				}
			}
			return cost;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	public double getFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		return 1;
	}

}
